import os
from abc import abstractmethod

class Grid:
    def __init__(self, metadata={}):
        self.metadata = metadata
        if self.metadata:
            self.data = []
            for r in range(self.metadata['rows']):
                self.data.append([None] * self.metadata['cols'])
        else:
            self.data = None
    
    def __str__(self):
        if not self.metadata:
            return "Grid not initialized"

        string = "Grid initialized by:" + os.linesep
        for k, v in self.metadata.items():
            string += f" {k}: {v}" + os.linesep
        string += "Data {}filled".format("not " if not self.data else "")
        
        return string

    def stats(self):
        minv = None
        maxv = None
        for row in self.data:
            for val in row:
                if minv is None or minv > val:
                    minv = val
                if maxv is None or maxv < val:
                    maxv = val

        return ({'min': minv, 'max': maxv})
                    
class DataGrid(Grid):
    def read_from_file(self, fd):
        # header
        row = fd.readline()
        while row:
            if not row[0].isnumeric():
                k, v = row.split(':')
                self.metadata[k] = int(v.strip())
            else:
                break
            row = fd.readline()
        if self.metadata['rows'] <= 0 or self.metadata['cols'] <= 0:
            raise Exception("Invalid grid size")

        # body
        self.data = []
        while row:
            self.data.append(list(map(lambda x: int(x), row.split())))
            if len(self.data[-1]) != self.metadata['cols']:
                raise Exception("Inconsistency in cols")
            row = fd.readline()
        if len(self.data) != self.metadata['rows']:
            raise Exception("Inconsistency in rows")
        
    def apply_kernel(self, kernel):
        output = DataGrid(self.metadata)

        for i in range(self.metadata['rows']):
            for j in range(self.metadata['cols']):
                kernel.reset()
                for l in range(kernel.metadata['rows']):
                    for k in range(kernel.metadata['cols']):
                        i_l = i + (l - kernel.size)
                        j_k = j + (k - kernel.size)
                        if i_l > -1 and j_k > -1 and \
                           i_l < self.metadata['rows'] and \
                           j_k < self.metadata['cols']:
                            kernel.data[l][k] = self.data[i_l][j_k]

                output.data[i][j] = kernel.calculate()
            if i > 0 and i % 10 == 0:
                perc = int(i / self.metadata['rows'] * 100)
                print(f"{perc}%", end="\r")
        print("100%")
        
        return output

class KernelGrid(Grid):
    def __init__(self, size):
        super().__init__({
            'rows' : size,
            'cols': size
        })
                
    @abstractmethod
    def calculate(self):
        pass

    def reset(self):
        for i in range(self.metadata['rows']):
            for j in range(self.metadata['cols']):
                self.data[i][j] = None

    @property
    def size(self):
        return int(self.metadata['rows'] / 2)
        
class KernelCount(KernelGrid):
    def calculate(self):
        count = 0
        for row in self.data:
            for val in row:
                if val is not None:
                    count += 1
    
        return count

class KernelSum(KernelGrid):
    def calculate(self):
        count = 0
        for row in self.data:
            for val in row:
                if val is not None:
                    count += val
    
        return count
    
