import os
from abc import abstractmethod

import numpy as np


class Grid:
    def __init__(self, metadata={}):
        self.metadata = metadata
        if self.metadata:
            self.data = np.empty(
                (self.metadata['rows'], self.metadata['cols']))
            self.data[:] = np.nan
        else:
            self.data = np.nan

    def __str__(self):
        if not self.metadata:
            return "Grid not initialized"

        string = "Grid initialized by:" + os.linesep
        for k, v in self.metadata.items():
            string += f" {k}: {v}" + os.linesep
        string += "Data {}filled".format("not " if not self.data.any() else "")

        return string

    def stats(self):
        return ({'min': self.data.min(), 'max': self.data.max()})


class DataGrid(Grid):
    def read_from_file(self, fd):
        # header
        row = fd.readline()
        while row:
            if not row[0].isnumeric():
                k, v = row.split(':')
                self.metadata[k] = int(v.strip())
            else:
                break
            row = fd.readline()
        if self.metadata['rows'] <= 0 or self.metadata['cols'] <= 0:
            raise Exception("Invalid grid size")

        # body
        array_1d = np.fromstring(fd.read(), sep='\n')
        self.data = np.resize(array_1d,
                              (self.metadata['rows'], self.metadata['cols']))

    def apply_kernel(self, kernel):
        output = DataGrid(self.metadata)
        buffered_data = np.empty(
            (self.metadata['rows'] + 2, self.metadata['cols'] + 2))
        buffered_data[:] = np.nan
        buffered_data[1:-1, 1:-1] = self.data.copy()

        for i in range(self.metadata['rows']):
            for j in range(self.metadata['cols']):
                kernel.data = buffered_data[i:i + kernel.metadata['rows'],
                                            j:j + kernel.metadata['cols']].copy()

                output.data[i][j] = kernel.calculate()
            if i > 0 and i % 10 == 0:
                perc = int(i / self.metadata['rows'] * 100)
                print(f"{perc}%", end="\r")
        print("100%")

        return output


class KernelGrid(Grid):
    def __init__(self, size):
        super().__init__({
            'rows': size,
            'cols': size
        })

    @abstractmethod
    def calculate(self):
        pass

    @property
    def size(self):
        return int(self.metadata['rows'] / 2)


class KernelCount(KernelGrid):
    def calculate(self):
        return self.data.size - np.count_nonzero(np.isnan(self.data))


class KernelSum(KernelGrid):
    def calculate(self):
        count = 0
        for row in self.data:
            for val in row:
                if val is not None:
                    count += val

        return count

